@@ -350,8 +350,11 @@ PROXY_DECLARE(request_rec *)ap_proxy_make_fake_req(conn_rec *c, request_rec *r)
     rp->status          = HTTP_OK;
 
     rp->headers_in      = apr_table_make(r->pool, 50);
+    rp->trailers_in     = apr_table_make(r->pool, 5);
+
     rp->subprocess_env  = apr_table_make(r->pool, 50);
     rp->headers_out     = apr_table_make(r->pool, 12);
+    rp->trailers_out    = apr_table_make(r->pool, 5);
     rp->err_headers_out = apr_table_make(r->pool, 5);
     rp->notes           = apr_table_make(r->pool, 5);
 
@@ -1644,6 +1647,14 @@ PROXY_DECLARE(int) ap_proxy_connect_to_backend(apr_socket_t **newsock,
     return connected ? 0 : 1;
 }
 
+PROXY_DECLARE(int) ap_proxy_connection_reusable(proxy_conn_rec *conn)
+{
+    proxy_worker *worker = conn->worker;
+
+    return ! (conn->close || conn->close_on_recycle
+              || !worker->is_address_reusable || worker->disablereuse);
+}
+
 static apr_status_t connection_cleanup(void *theconn)
 {
     proxy_conn_rec *conn = (proxy_conn_rec *)theconn;
@@ -1669,8 +1680,7 @@ static apr_status_t connection_cleanup(void *theconn)
 #endif
 
     /* determine if the connection need to be closed */
-    if (conn->close_on_recycle || conn->close || worker->disablereuse ||
-        !worker->is_address_reusable) {
+    if (!ap_proxy_connection_reusable(conn)) {
         apr_pool_t *p = conn->pool;
         apr_pool_clear(p);
         conn = apr_pcalloc(p, sizeof(proxy_conn_rec));
@@ -1698,6 +1708,7 @@ static void socket_cleanup(proxy_conn_rec *conn)
 {
     conn->sock = NULL;
     conn->connection = NULL;
+    conn->ssl_hostname = NULL;
     apr_pool_clear(conn->scpool);
 }
 
@@ -1790,6 +1801,76 @@ static apr_status_t connection_destructor(void *resource, void *params,
 }
 #endif
 
+#if PROXY_HAS_SCOREBOARD
+PROXY_DECLARE(void*) ap_proxy_set_scoreboard_lb(proxy_worker *worker,
+                                                proxy_balancer *balancer,
+                                                server_rec *server)
+{
+    if (ap_scoreboard_image && !worker->s) {
+        int i = 0;
+        proxy_worker_stat *free_slot = NULL;
+        proxy_worker_stat *s;
+        unsigned char digest[APR_MD5_DIGESTSIZE];
+        apr_md5_ctx_t ctx;
+
+        /* The scoreboard entry must be unique per server and balancer,
+         * so when asked to (given by the caller) use their respective
+         * unique identifiers for the hash.
+         */
+        apr_md5_init(&ctx);
+        apr_md5_update(&ctx, (unsigned char *)worker->name,
+                       strlen(worker->name));
+        if (balancer) {
+            apr_md5_update(&ctx, (unsigned char *)balancer->name,
+                           strlen(balancer->name));
+        }
+        if (server) {
+            server_addr_rec *addr;
+            /* Assumes the unique identifier of a vhost is its address(es)
+             * plus the ServerName:Port. Should two or more vhosts have this
+             * same identifier, the first one would always be elected to
+             * handle the requests, so this shouldn't be an issue...
+             */
+            for (addr = server->addrs; addr; addr = addr->next) {
+                char host_ip[64]; /* for any IPv[46] string */
+                apr_sockaddr_ip_getbuf(host_ip, sizeof host_ip,
+                                       addr->host_addr);
+                apr_md5_update(&ctx, (unsigned char *)host_ip,
+                               strlen(host_ip));
+                apr_md5_update(&ctx, (unsigned char *)&addr->host_port,
+                               sizeof(addr->host_port));
+            }
+            apr_md5_update(&ctx, (unsigned char *)server->server_hostname,
+                           strlen(server->server_hostname));
+            apr_md5_update(&ctx, (unsigned char *)&server->port,
+                           sizeof(server->port));
+        }
+        apr_md5_final(digest, &ctx);
+
+        /* Try to find out the right shared memory according to the hash */
+        while ((s = (proxy_worker_stat *)ap_get_scoreboard_lb(i++)) != NULL) {
+            if ((s->status & PROXY_WORKER_INITIALIZED) == 0) {
+                if (free_slot == NULL) {
+                    free_slot = s;
+                }
+                continue;
+            }
+            if (memcmp(s->digest, digest, APR_MD5_DIGESTSIZE) == 0) {
+                worker->s = s;
+                return s;
+            }
+        }
+
+        /* We failed to find out shared memory, so just use a free slot (if any) */
+        if (free_slot) {
+            memcpy(free_slot->digest, digest, APR_MD5_DIGESTSIZE);
+            worker->s = free_slot;
+        }
+    }
+    return worker->s;
+}
+#endif
+
 /*
  * ap_proxy_initialize_worker_share() concerns itself
  * with initializing those parts of worker which
@@ -1799,11 +1880,7 @@ PROXY_DECLARE(void) ap_proxy_initialize_worker_share(proxy_server_conf *conf,
                                                      proxy_worker *worker,
                                                      server_rec *s)
 {
-#if PROXY_HAS_SCOREBOARD
-    lb_score *score = NULL;
-#else
     void *score = NULL;
-#endif
 
     if (PROXY_WORKER_IS_INITIALIZED(worker)) {
         /* The worker share is already initialized */
@@ -1813,18 +1890,18 @@ PROXY_DECLARE(void) ap_proxy_initialize_worker_share(proxy_server_conf *conf,
         return;
     }
 #if PROXY_HAS_SCOREBOARD
-        /* Get scoreboard slot */
+    /* Get scoreboard slot */
     if (ap_scoreboard_image) {
-        score = ap_get_scoreboard_lb(worker->id);
-        if (!score) {
+        if (!ap_proxy_set_scoreboard_lb(worker, NULL, s)) {
             ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-                  "proxy: ap_get_scoreboard_lb(%d) failed in child %" APR_PID_T_FMT " for worker %s",
+                  "proxy: ap_proxy_set_scoreboard_lb(%d) failed in child %" APR_PID_T_FMT " for worker %s",
                   worker->id, getpid(), worker->name);
         }
         else {
              ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                   "proxy: grabbed scoreboard slot %d in child %" APR_PID_T_FMT " for worker %s",
                   worker->id, getpid(), worker->name);
+             score = worker->s;
         }
     }
 #endif
@@ -1833,8 +1910,8 @@ PROXY_DECLARE(void) ap_proxy_initialize_worker_share(proxy_server_conf *conf,
         ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
               "proxy: initialized plain memory in child %" APR_PID_T_FMT " for worker %s",
               getpid(), worker->name);
+        worker->s = (proxy_worker_stat *)score;
     }
-    worker->s = (proxy_worker_stat *)score;
     /*
      * recheck to see if we've already been here. Possible
      * if proxy is using scoreboard to hold shared stats
@@ -1860,7 +1937,6 @@ PROXY_DECLARE(void) ap_proxy_initialize_worker_share(proxy_server_conf *conf,
     }
 
     worker->s->status |= (worker->status | PROXY_WORKER_INITIALIZED);
-
 }
 
 PROXY_DECLARE(apr_status_t) ap_proxy_initialize_worker(proxy_worker *worker, server_rec *s)
@@ -2080,8 +2156,12 @@ ap_proxy_determine_connection(apr_pool_t *p, request_rec *r,
      * The scheme handler decides if this is permanent or
      * short living pool.
      */
-    /* are we connecting directly, or via a proxy? */
-    if (!proxyname) {
+    /* Unless we are connecting the backend via a (forward Proxy)Remote, we
+     * have to use the original form of the URI (non absolute), but this is
+     * also the case via a remote proxy using the CONNECT method since the
+     * original request (and URI) is to be embedded in the body.
+     */
+    if (!proxyname || conn->is_ssl) {
         *url = apr_pstrcat(p, uri->path, uri->query ? "?" : "",
                            uri->query ? uri->query : "",
                            uri->fragment ? "#" : "",
@@ -2197,6 +2277,38 @@ ap_proxy_determine_connection(apr_pool_t *p, request_rec *r,
         return ap_proxyerror(r, HTTP_FORBIDDEN,
                              "Connect to remote machine blocked");
     }
+    /*
+     * When SSL is configured, determine the hostname (SNI) for the request
+     * and save it in conn->ssl_hostname. Close any reused connection whose
+     * SNI differs.
+     */
+    if (conn->is_ssl) {
+        const char *ssl_hostname;
+        /*
+         * In the case of ProxyPreserveHost on use the hostname of
+         * the request if present otherwise use the one from the
+         * backend request URI.
+         */
+        if (conf->preserve_host) {
+            ssl_hostname = r->hostname;
+        }
+        else {
+            ssl_hostname = conn->hostname;
+        }
+        /*
+         * Close if a SNI is in use but this request requires no or
+         * a different one, or no SNI is in use but one is required.
+         */
+        if ((conn->ssl_hostname && (!ssl_hostname ||
+                                    strcasecmp(conn->ssl_hostname,
+                                               ssl_hostname) != 0)) ||
+                (!conn->ssl_hostname && ssl_hostname && conn->sock)) {
+            socket_cleanup(conn);
+        }
+        if (conn->ssl_hostname == NULL) {
+            conn->ssl_hostname = apr_pstrdup(conn->scpool, ssl_hostname);
+        }
+    }
     ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
                  "proxy: connected %s to %s:%d", *url, conn->hostname,
                  conn->port);
@@ -2329,7 +2441,7 @@ static apr_status_t send_http_connect(proxy_conn_rec *backend,
         nbytes = sizeof(drain_buffer) - 1;
         while (status == APR_SUCCESS && nbytes) {
             status = apr_socket_recv(backend->sock, drain_buffer, &nbytes);
-            buffer[nbytes] = '\0';
+            drain_buffer[nbytes] = '\0';
             nbytes = sizeof(drain_buffer) - 1;
             if (strstr(drain_buffer, "\r\n\r\n") != NULL) {
                 complete = 1;
